this.asyncStorage = function () {
    'use strict';
    var DBNAME = 'asyncStorage';
    var DBVERSION = 1;
    var STORENAME = 'keyvaluepairs';
    var db = null;
    function withDatabase(f) {
        if (db) {
            f();
        } else {
            var openreq = indexedDB.open(DBNAME, DBVERSION);
            openreq.onerror = function withStoreOnError() {
                console.error('asyncStorage: can\'t open database:', openreq.error.name);
            };
            openreq.onupgradeneeded = function withStoreOnUpgradeNeeded() {
                openreq.result.createObjectStore(STORENAME);
            };
            openreq.onsuccess = function withStoreOnSuccess() {
                db = openreq.result;
                f();
            };
        }
    }
    function withStore(type, callback, oncomplete) {
        withDatabase(function () {
            var transaction = db.transaction(STORENAME, type);
            if (oncomplete) {
                transaction.oncomplete = oncomplete;
            }
            callback(transaction.objectStore(STORENAME));
        });
    }
    function getItem(key, callback) {
        var req;
        withStore('readonly', function getItemBody(store) {
            req = store.get(key);
            req.onerror = function getItemOnError() {
                console.error('Error in asyncStorage.getItem(): ', req.error.name);
            };
        }, function onComplete() {
            var value = req.result;
            if (value === undefined) {
                value = null;
            }
            callback(value);
        });
    }
    function setItem(key, value, callback) {
        withStore('readwrite', function setItemBody(store) {
            var req = store.put(value, key);
            req.onerror = function setItemOnError() {
                console.error('Error in asyncStorage.setItem(): ', req.error.name);
            };
        }, callback);
    }
    function removeItem(key, callback) {
        withStore('readwrite', function removeItemBody(store) {
            var req = store.delete(key);
            req.onerror = function removeItemOnError() {
                console.error('Error in asyncStorage.removeItem(): ', req.error.name);
            };
        }, callback);
    }
    function clear(callback) {
        withStore('readwrite', function clearBody(store) {
            var req = store.clear();
            req.onerror = function clearOnError() {
                console.error('Error in asyncStorage.clear(): ', req.error.name);
            };
        }, callback);
    }
    function length(callback) {
        var req;
        withStore('readonly', function lengthBody(store) {
            req = store.count();
            req.onerror = function lengthOnError() {
                console.error('Error in asyncStorage.length(): ', req.error.name);
            };
        }, function onComplete() {
            callback(req.result);
        });
    }
    function key(n, callback) {
        if (n < 0) {
            callback(null);
            return;
        }
        var req;
        withStore('readonly', function keyBody(store) {
            var advanced = false;
            req = store.openCursor();
            req.onsuccess = function keyOnSuccess() {
                var cursor = req.result;
                if (!cursor) {
                    return;
                }
                if (n === 0 || advanced) {
                    return;
                }
                advanced = true;
                cursor.advance(n);
            };
            req.onerror = function keyOnError() {
                console.error('Error in asyncStorage.key(): ', req.error.name);
            };
        }, function onComplete() {
            var cursor = req.result;
            callback(cursor ? cursor.key : null);
        });
    }
    return {
        getItem: getItem,
        setItem: setItem,
        removeItem: removeItem,
        clear: clear,
        length: length,
        key: key
    };
}();